#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include "misc.h"
#include <QProcess>

#include <linux/dvb/frontend.h>
#include <linux/dvb/dmx.h>

#include "zap.h"

extern bool do_zap;
extern int update_ms;
extern int numAdapters;
//extern bool useLastFirst;
extern char chip_name[1024];
extern int specific_adapter; // = -1;
extern char specific_chip[128];
extern int current_adapter;

static char FRONTEND_DEV [80];
static char DEMUX_DEV [80];
char ERRBUFF[256];

const char *type_str[] = {
	"QPSK",
	"QAM",
	"OFDM",
	"ATSC",
};

const char *del_str[] = {
	"UNDEFINED",
	"DVB-C (A)",
	"DVB-C (B)",
	"DVB-T",
	"DSS",
	"DVB-S",
	"DVB-S2",
	"DVB-H",
	"ISDB-T",
	"ISDB-S",
	"ISDB-C",
	"ATSC",
	"ATSC-M/H",
	"DTMB",
	"CMMB",
	"DAB",
	"DVB-T2",
	"TURBO",
	"QAM (C)",
};


int set_pesfilter(int dmxfd, int pid, int pes_type)
{
	struct dmx_pes_filter_params pesfilter;

	/* ignore this pid to allow radio services */
	if (pid < 0 || pid >= 0x1fff || (pid == 0 && pes_type != DMX_PES_OTHER))
		return 0;

	//if (dvr) {
		int buffersize = 64 * 1024;
		if (ioctl(dmxfd, DMX_SET_BUFFER_SIZE, buffersize) == -1)
			perror("DMX_SET_BUFFER_SIZE failed");
	//}

	pesfilter.pid = pid;
	pesfilter.input = DMX_IN_FRONTEND;
	//pesfilter.output = dvr ? DMX_OUT_TS_TAP : DMX_OUT_DECODER;
	pesfilter.output = DMX_OUT_TS_TAP;
	//pesfilter.output = DMX_OUT_DECODER;
	pesfilter.pes_type = (dmx_ts_pes) pes_type;
	pesfilter.flags = DMX_IMMEDIATE_START;

	if (ioctl(dmxfd, DMX_SET_PES_FILTER, &pesfilter) == -1) {
		snprintf(ERRBUFF, 255, "DMX_SET_PES_FILTER failed (PID = 0x%04x): %d %m\n", pid, errno);
		return -1;
	}
	return 0;
}

/* to be used with v3 drivers */
int check_frontend_v3(int fd, enum fe_type type)
{
	struct dvb_frontend_info info;
	int ret;

	ret = ioctl(fd, FE_GET_INFO, &info);
	if (ret < 0) {
		snprintf(ERRBUFF, 255, "ioctl FE_GET_INFO failed\n");
		close(fd);
		ret = -1;
		goto exit;
	}
	if (info.type != type) {
		snprintf(ERRBUFF, 255, "Not a valid %s device!\n", type_str[type]);
		close(fd);
		ret = -EINVAL;
		goto exit;
	}
exit:
	return ret;
}

static int map_delivery_mode(fe_type_t *type, enum fe_delivery_system delsys)
{
	switch (delsys) {
	case SYS_DSS:
	case SYS_DVBS:
	case SYS_DVBS2:
	case SYS_TURBO:
		*type = FE_QPSK;
		break;
	case SYS_DVBT:
	case SYS_DVBT2:
	case SYS_DVBH:
	case SYS_ISDBT:
		*type = FE_OFDM;
		break;
	case SYS_DVBC_ANNEX_A:
	case SYS_DVBC_ANNEX_C:
		*type = FE_QAM;
		break;
	case SYS_ATSC:
	case SYS_DVBC_ANNEX_B:
		*type = FE_ATSC;
		break;
	default:
		snprintf(ERRBUFF, 255, "Delivery system unsupported, please report to linux-media ML\n");
		return -1;
	}
	return 0;
}

int get_property(int fd, uint32_t pcmd, uint32_t *len, uint8_t *data)
{
	struct dtv_property p, *b;
	struct dtv_properties cmd;
	int ret;

	p.cmd = pcmd;
	cmd.num = 1;
	cmd.props = &p;
	b = &p;

	ret = ioctl(fd, FE_GET_PROPERTY, &cmd);
	if (ret < 0) {
		snprintf(ERRBUFF, 255, "FE_SET_PROPERTY returned %d\n", ret);
		return -1;
	}
	memcpy(len, &b->u.buffer.len, sizeof (uint32_t));
	memcpy(data, b->u.buffer.data, *len);
	return 0;
}

int set_property(int fd, uint32_t cmd, uint32_t data)
{
	struct dtv_property p, *b;
	struct dtv_properties c;
	int ret;

	p.cmd = cmd;
	c.num = 1;
	c.props = &p;
	b = &p;
	b->u.data = data;
	ret = ioctl(fd, FE_SET_PROPERTY, &c);
	if (ret < 0) {
		snprintf(ERRBUFF, 255, "FE_SET_PROPERTY returned %d\n", ret);
		return -1;
	}
	return 0;
}

int dvbfe_get_delsys(int fd, fe_delivery_system_t *delsys)
{
	uint32_t len;
	/* Buggy API design */
	return get_property(fd, DTV_DELIVERY_SYSTEM, &len, (uint8_t *)delsys);
}

int dvbfe_set_delsys(int fd, enum fe_delivery_system delsys)
{
	return set_property(fd, DTV_DELIVERY_SYSTEM, delsys);
}

int dvbfe_enum_delsys(int fd, uint32_t *len, uint8_t *data)
{
	return get_property(fd, DTV_ENUM_DELSYS, len, data);
}

int dvbfe_get_version(int fd, int *major, int *minor)
{
	struct dtv_property p, *b;
	struct dtv_properties cmd;
	int ret;

	p.cmd = DTV_API_VERSION;
	cmd.num = 1;
	cmd.props = &p;
	b = &p;

	ret = ioctl(fd, FE_GET_PROPERTY, &cmd);
	if (ret < 0) {
		snprintf(ERRBUFF, 255, "FE_GET_PROPERTY failed, ret=%d\n", ret);
		return -1;
	}
	*major = (b->u.data >> 8) & 0xff;
	*minor = b->u.data & 0xff;
	return 0;
}

int check_frontend_multi(int fd, enum fe_type type, uint32_t *mstd)
{
	int ret;

	enum fe_type delmode;
	unsigned int i, valid_delsys = 0;
	uint32_t len;
	uint8_t data[32];

	ret = dvbfe_enum_delsys(fd, &len, data);
	if (ret) {
		snprintf(ERRBUFF, 255, "enum_delsys failed, ret=%d\n", ret);
		ret = -EIO;
		goto exit;
	}
	snprintf(ERRBUFF, 255, "\t FE_CAN { ");
	for (i = 0; i < len; i++) {
		if (i < len - 1)
			snprintf(ERRBUFF, 255, "%s + ", del_str[data[i]]);
		else
			snprintf(ERRBUFF, 255, "%s", del_str[data[i]]);
	}
	snprintf(ERRBUFF, 255, " }\n");
	/* check whether frontend can support our delivery */
	for (i = 0; i < len; i++) {
		map_delivery_mode(&delmode, (fe_delivery_system) data[i]);
		if (type == delmode) {
			valid_delsys = 1;
			ret = 0;
			break;
		}
	}
	if (!valid_delsys) {
		snprintf(ERRBUFF, 255, "Not a valid %s device!\n", type_str[type]);
		ret = -EINVAL;
		goto exit;
	}
	*mstd = len; /* mstd has supported delsys count */
exit:
	return ret;
}

int check_frontend(int fd, enum fe_type type, uint32_t *mstd)
{
	int major, minor, ret;

	ret = dvbfe_get_version(fd, &major, &minor);
	if (ret)
		goto exit;
	snprintf(ERRBUFF, 255, "Version: %d.%d  ", major, minor);
	if ((major == 5) && (minor > 8)) {
		ret = check_frontend_multi(fd, type, mstd);
		if (ret)
			goto exit;
	} else {
		ret = check_frontend_v3(fd, type);
		if (ret)
			goto exit;
	}
exit:
	return ret;
}

static
int setup_frontend (int fe_fd, struct dvb_frontend_parameters *frontend)
{
	uint32_t mstd;

	if (check_frontend(fe_fd, FE_ATSC, &mstd) < 0) {
		close(fe_fd);
		return -1;
	}

	/* TODO! Some frontends need to be explicit delivery system */
	//printf ("tuning to %i Hz\n", frontend->frequency);

	if (ioctl(fe_fd, FE_SET_FRONTEND, frontend) < 0) {
		snprintf(ERRBUFF, 255, "ioctl FE_SET_FRONTEND failed");
		return -1;
	}

	return 0;
}

// this constructor is run in MainWindow's thread
Zapper::Zapper(unsigned int Freq, int Apid, int Vpid, int Sid)
{
	freq = Freq;
	apid = Apid;
	vpid = Vpid;
	sid = Sid;
}

// This function called by MainWindow and, like Zapper::Zapper(), is run in that object's thread.
// The sendInfo() signal, therefore, is a direct connection so the MainWindow::showInfo() slot is
// finished before the infubuf[] disappears at the end of this function.
bool Zapper::open_frontend()
{
	struct dvb_frontend_parameters frontend_param;
	int frontend = 0, demux = 0;
	char infobuf[1024];

	current_adapter = -1;
	ready = false;
	zapping = false;
	memset(ERRBUFF, 0, 256);

	int num_adapters = 0;

	FILE * cmd = popen("/bin/ls /dev/dvb | wc -l", "r");
	fscanf(cmd, "%d", &num_adapters);
	pclose(cmd);

	if (numAdapters > 0 && num_adapters != numAdapters)
		num_adapters = numAdapters;

	// used to have a checkbox that used last adapter first.  Changed structure to be able to specify chip or adapter
/*
	bool useLastFirst = false;
	for (adapter = (numAdapters < 0 || useLastFirst) ? num_adapters-1 : 0;
	 (numAdapters >= 0 && !useLastFirst && adapter < num_adapters) || ((numAdapters < 0 || useLastFirst) && adapter >= 0);
	 adapter += (numAdapters < 0 || useLastFirst) ? -1 : 1) {
*/
	for (adapter = 0; adapter < num_adapters; adapter++) {

		if (specific_adapter >= 0 && adapter != specific_adapter)
			continue; 

		// Get the adapter's chip name
		// dmesg | grep 'registering adapter'    
		// [   16.797196] cx23885 0000:0b:00.0: DVB: registering adapter 0 frontend 0 (Samsung S5H1411 QAM/8VSB Frontend)...
		// [   16.849681] cx23885 0000:0b:00.0: DVB: registering adapter 1 frontend 0 (Samsung S5H1411 QAM/8VSB Frontend)...
		// [281448.513000] em28xx 3-1:1.0: DVB: registering adapter 2 frontend 0 (LG Electronics LGDT3306A VSB/QAM Frontend)...
		// [281448.522076] em28xx 3-1:1.0: DVB: registering adapter 3 frontend 0 (LG Electronics LGDT3306A VSB/QAM Frontend)...
		memset(chip_name, 0, 1024);
		char dmsg_cmd[2048];
		sprintf(dmsg_cmd, "dmesg | sed -n -r 's/\\[[ 0-9.]+\\]\\s+(\\S+)\\s+\\S+\\s+DVB: registering adapter %d frontend [0-9] \\((\\S+) .+/\\1_\\2/p'", adapter);
	
		FILE * dmesg = popen(dmsg_cmd, "r");
		if (dmesg) {
			fgets(chip_name, 1023, dmesg);
			pclose(dmesg);
		}

		int cnlen = strlen(chip_name);

		// See if there is a saved copy of dmesg that contains the chip
		if (cnlen < 3) {
			fprintf(stderr, "Checking /tmp/dmesg.out for chip name\n");
			sprintf(dmsg_cmd, "cat /tmp/dmesg.out | sed -n -r 's/\\[[ 0-9.]+\\]\\s+(\\S+)\\s+\\S+\\s+DVB: registering adapter %d frontend [0-9] \\((\\S+) .+/\\1_\\2/p'", adapter);
			FILE * tmpdmesg = popen(dmsg_cmd, "r");
			if (tmpdmesg) {
				fgets(chip_name, 1023, dmesg);
				pclose(tmpdmesg);
				cnlen = strlen(chip_name);
			}
		}

		if (cnlen < 3)
			strcpy(chip_name, "unknown");
		else
			chip_name[cnlen-1] = '\0'; // remove trailing \n

		if (strlen(specific_chip) && strcmp(chip_name, specific_chip))
			continue;
	
		snprintf (FRONTEND_DEV, sizeof(FRONTEND_DEV), "/dev/dvb/adapter%i/frontend%i", adapter, frontend);

		snprintf (DEMUX_DEV, sizeof(DEMUX_DEV), "/dev/dvb/adapter%i/demux%i", adapter, demux);

		snprintf(infobuf, 1023, "Opening adapter %s (%s)", FRONTEND_DEV, chip_name); //, DEMUX_DEV);
		emit sendInfo(infobuf);
		usleep(10000);

		if ((frontend_fd = open(FRONTEND_DEV, O_RDWR | O_NONBLOCK)) < 0) {
			if (errno == EBUSY)
				snprintf(infobuf, 1023, "adapter %d in use\n", adapter);
			else
				snprintf(infobuf, 1023, "Error %d opening adapter %d: %s\n", errno, adapter, strerror(errno));
			emit sendInfo(infobuf);
			usleep(10000);
		} // frontend_fd < 0
		else
			break;
	}

	// no luck
	if (frontend_fd < 0) {
		emit sendInfo("No adapters available");
		usleep(10000);
		return true;
	}

	memset(&frontend_param, 0, sizeof(struct dvb_frontend_parameters));

	//if (parse (confname, channel, &frontend_param, &vpid, &apid, &sid))
	//	return -1;
	frontend_param.frequency = freq;
	frontend_param.u.vsb.modulation = VSB_8;

	if (setup_frontend (frontend_fd, &frontend_param) < 0) {
		snprintf(infobuf, 1023, "Error setting up frontend: %s", ERRBUFF);
		emit sendInfo(infobuf);
		usleep(10000);
		return true;
	}

	if ((video_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
		snprintf(infobuf, 1023, "Error opening %s: %s", DEMUX_DEV, ERRBUFF);
		emit sendInfo(infobuf);
		usleep(10000);
		close (frontend_fd);
		return true;
	}

	if (set_pesfilter (video_fd, vpid, DMX_PES_VIDEO) < 0) {
		snprintf(infobuf, 1023, "Error setting PES filter: %s", ERRBUFF);
		emit sendInfo(infobuf);
		usleep(10000);
		close (frontend_fd);
		return true;
	}

	if ((audio_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
		snprintf(infobuf, 1023, "Error opening audio PID on %s: %s\n", DEMUX_DEV, ERRBUFF);
		emit sendInfo(infobuf);
		usleep(10000);
		close (video_fd);
		close (frontend_fd);
		return true;
	}

	if (set_pesfilter (audio_fd, apid, DMX_PES_AUDIO) < 0) {
		snprintf(infobuf, 1023, "Error setting audio PES filter: %s", ERRBUFF);
		emit sendInfo(infobuf);
		usleep(10000);
		close (video_fd);
		close (frontend_fd);
		return true;
	}

	ready = true;
	current_adapter = adapter;

//	emit sendInfo("Frontend opened"); // Ready to start monitoring\n");
//	usleep(10000);
	
	return false;
} // open_frontend()

// This is the background thread
void Zapper::run(void)
{
	fe_status_t status;
	uint16_t snr, signal;
	//uint32_t ber, uncorrected_blocks;

	zapping = true;
	
	// This emit caused a GPF when monitoring stopped.  WHY!!?!
	//	emit sendInfo("Channel tuned, monitoring..."); // Ready to start monitoring\n");

	while (do_zap) {
		ioctl(frontend_fd, FE_READ_STATUS, &status);
		ioctl(frontend_fd, FE_READ_SIGNAL_STRENGTH, &signal);
		ioctl(frontend_fd, FE_READ_SNR, &snr);

		emit newReading(((float) signal / 65536.0) * 1e2, snr, status & FE_HAS_LOCK);;

		usleep(update_ms * 1000);

	} // while do_zap

	close (audio_fd);
	close (video_fd);
	close (frontend_fd);

	zapping = false;
}


